using System;
using System.Collections.Generic;
using X3g.Plugin;
using X3g.Plugin.Modeling;

namespace SamplePlugin.Create
{
    class Modeling
    {
        private DocumentManager mDocument;

        public Modeling(DocumentManager pDocument)
        {
            mDocument = pDocument;
        }
        public void CreateDiceTable()
        {
            var tTable = CreateTable();
            tTable.UpdateGeometry();

            var tDice = CreateDice();
            tDice.TransformBy(GeMatrix.Translate(0, 0, 1.0));
            tDice.UpdateGeometry();

            var tDiceCup = CreateDiceCup();
            tDiceCup.TransformBy(GeMatrix.Translate(-0.02, 0, 0)
                .PostMult(GeMatrix.Rotate(Math.Atan(0.06 / 0.01), 0, 1, 0))
                .PostMult(GeMatrix.Rotate(Math.PI / 4, 0, 0, -1))
                .PostMult(GeMatrix.Translate(-0.05, 0.05, 1)));
            tDiceCup.UpdateGeometry();
        }

        private DbEntity CreateTable()
        {
            var tTableBlock = mDocument.Blocks.CreateBlock("");
            var tTableTop = CreateTableTop(tTableBlock);

            // Create Table legs
            var tLegBlock = mDocument.Blocks.CreateBlock("");
            var tTableLeg = Csg.CreateCone(new GeVec3d(), 0.02, new GeVec3d(0, 0, 0.99), 0.02, tLegBlock);
            tTableLeg.Color = new EntityColor(System.Drawing.Color.Silver);

            var tLegRef1 = new DbBlockReference(tTableBlock, tLegBlock);
            tLegRef1.Position = new GeVec3d(0.4, 0.2, 0);
            tLegRef1.UpdateGeometry();

            var tLegRef2 = new DbBlockReference(tTableBlock, tLegBlock);
            tLegRef2.Position = new GeVec3d(0.4, -0.2, 0);
            tLegRef2.UpdateGeometry();

            var tLegRef3 = new DbBlockReference(tTableBlock, tLegBlock);
            tLegRef3.Position = new GeVec3d(-0.4, 0.2, 0);
            tLegRef3.UpdateGeometry();

            var tLegRef4 = new DbBlockReference(tTableBlock, tLegBlock);
            tLegRef4.Position = new GeVec3d(-0.4, -0.2, 0);
            tLegRef4.UpdateGeometry();

            return new DbBlockReference(mDocument.ModelSpace, tTableBlock);
        }

        private DbEntity CreateTableTop(Block pOwner)
        {
            GeVec3d[] tVertices = new GeVec3d[2];

            double tRound = 0.01;
            GeVec2d tOuter = new GeVec2d(0.5, 0.25);
            GeVec2d tInner = new GeVec2d(tOuter.x - tRound, tOuter.y - tRound);

            List<DbCurve> tContour = new List<DbCurve>();
            {
                // Back
                var tLine = new DbPolyline(mDocument.ModelSpace);
                tVertices[0] = new GeVec3d(tInner.x, tOuter.y, 1);
                tVertices[1] = new GeVec3d(-tInner.x, tOuter.y, 1);
                tLine.SetVertices(tVertices);

                tContour.Add(tLine);

                var tEllipse = new DbEllipse(mDocument.ModelSpace);
                tEllipse.Normal = new GeVec3d(0, 0, 1);
                tEllipse.Center = new GeVec3d(-tInner.x, tInner.y, 1);
                tEllipse.MajorAxis = new GeVec3d(0, tRound, 0);
                tEllipse.RadiusRatio = 1;
                tEllipse.StartAngle = 0;
                tEllipse.EndAngle = Math.PI / 2;

                tContour.Add(tEllipse);
            }
            {
                // Left
                var tLine = new DbPolyline(mDocument.ModelSpace);
                tVertices[0] = new GeVec3d(-tOuter.x, tInner.y, 1);
                tVertices[1] = new GeVec3d(-tOuter.x, -tInner.y, 1);
                tLine.SetVertices(tVertices);

                tContour.Add(tLine);

                var tEllipse = new DbEllipse(mDocument.ModelSpace);
                tEllipse.Normal = new GeVec3d(0, 0, 1);
                tEllipse.Center = new GeVec3d(-tInner.x, -tInner.y, 1);
                tEllipse.MajorAxis = new GeVec3d(-tRound, 0, 0);
                tEllipse.RadiusRatio = 1;
                tEllipse.StartAngle = 0;
                tEllipse.EndAngle = Math.PI / 2;

                tContour.Add(tEllipse);
            }
            {
                // Front
                var tLine = new DbPolyline(mDocument.ModelSpace);
                tVertices[0] = new GeVec3d(-tInner.x, -tOuter.y, 1);
                tVertices[1] = new GeVec3d(tInner.x, -tOuter.y, 1);
                tLine.SetVertices(tVertices);

                tContour.Add(tLine);

                var tEllipse = new DbEllipse(mDocument.ModelSpace);
                tEllipse.Normal = new GeVec3d(0, 0, 1);
                tEllipse.Center = new GeVec3d(tInner.x, -tInner.y, 1);
                tEllipse.MajorAxis = new GeVec3d(0, -tRound, 0);
                tEllipse.RadiusRatio = 1;
                tEllipse.StartAngle = 0;
                tEllipse.EndAngle = Math.PI / 2;

                tContour.Add(tEllipse);
            }
            {
                // Right
                var tLine = new DbPolyline(mDocument.ModelSpace);
                tVertices[0] = new GeVec3d(tOuter.x, -tInner.y, 1);
                tVertices[1] = new GeVec3d(tOuter.x, tInner.y, 1);
                tLine.SetVertices(tVertices);

                tContour.Add(tLine);

                var tEllipse = new DbEllipse(mDocument.ModelSpace);
                tEllipse.Normal = new GeVec3d(0, 0, 1);
                tEllipse.Center = new GeVec3d(tInner.x, tInner.y, 1);
                tEllipse.MajorAxis = new GeVec3d(tRound, 0, 0);
                tEllipse.RadiusRatio = 1;
                tEllipse.StartAngle = 0;
                tEllipse.EndAngle = Math.PI / 2;

                tContour.Add(tEllipse);
            }

            var tTableTop = Csg.Extrude(tContour, new GeVec3d(0, 0, -0.01), pOwner);
            tTableTop.Color = new EntityColor(System.Drawing.Color.Tan);
            tTableTop.UpdateGeometry();

            foreach (var tCurve in tContour)
                tCurve.Delete();

            return tTableTop;
        }

        private DbEntity CreateDice()
        {
            double tDotD = 0.45;
            var tDice = Csg.CreateBox(new GeVec3d(-1, -1, -1), new GeVec3d(1, 1, 1), mDocument.ModelSpace);
            tDice.Color = new EntityColor(System.Drawing.Color.White);

            var tSphere = Csg.CreateSphere(new GeVec3d(), 1.35, mDocument.ModelSpace);
            var tNewDice = tDice.Common2(tSphere);
            tSphere.Delete();
            tDice.Delete();
            tDice = tNewDice;

            List<GeVec3d> tPoints = new List<GeVec3d>();
            // Bottom (Show a one)
            tPoints.Add(new GeVec3d(0, 0, -1));
            // Left (Show a two)
            tPoints.Add(new GeVec3d(-1, tDotD, tDotD));
            tPoints.Add(new GeVec3d(-1, -tDotD, -tDotD));
            // Back (Show a three)
            tPoints.Add(new GeVec3d(tDotD, 1, tDotD));
            tPoints.Add(new GeVec3d(0, 1, 0));
            tPoints.Add(new GeVec3d(-tDotD, 1, -tDotD));
            // Front (Show a four)
            tPoints.Add(new GeVec3d(tDotD, -1, tDotD));
            tPoints.Add(new GeVec3d(tDotD, -1, -tDotD));
            tPoints.Add(new GeVec3d(-tDotD, -1, tDotD));
            tPoints.Add(new GeVec3d(-tDotD, -1, -tDotD));
            // Right (Show a five)
            tPoints.Add(new GeVec3d(1, tDotD, tDotD));
            tPoints.Add(new GeVec3d(1, tDotD, -tDotD));
            tPoints.Add(new GeVec3d(1, 0, 0));
            tPoints.Add(new GeVec3d(1, -tDotD, tDotD));
            tPoints.Add(new GeVec3d(1, -tDotD, -tDotD));
            // Top (Show a six)
            tPoints.Add(new GeVec3d(tDotD, tDotD, 1));
            tPoints.Add(new GeVec3d(tDotD, 0, 1));
            tPoints.Add(new GeVec3d(tDotD, -tDotD, 1));
            tPoints.Add(new GeVec3d(-tDotD, tDotD, 1));
            tPoints.Add(new GeVec3d(-tDotD, 0, 1));
            tPoints.Add(new GeVec3d(-tDotD, -tDotD, 1));

            var tLastPos = new GeVec3d();
            var tHole = Csg.CreateSphere(tLastPos, 0.18, mDocument.ModelSpace);

            foreach (var tNextPoint in tPoints)
            {
                tHole.TransformBy(GeMatrix.Translate(tNextPoint - tLastPos));
                tNewDice = tDice.Cut2(tHole);
                tDice.Delete();
                tDice = tNewDice;
                tLastPos = tNextPoint;
            }

            tHole.Delete();

            tDice.TransformBy(GeMatrix.Scale(0.01).PostMult(GeMatrix.Translate(0, 0, 0.01)));

            return tDice;
        }

        private DbEntity CreateDiceCup()
        {
            double tThickness = 0.001;

            var tCone = Csg.CreateCone(new GeVec3d(0, 0, 0), 0.02, new GeVec3d(0, 0, 0.06), 0.03, mDocument.ModelSpace);
            var tCutOut = Csg.CreateCone(new GeVec3d(0, 0, tThickness), 0.02 - tThickness / 2, new GeVec3d(0, 0, 0.06), 0.03 - tThickness / 2, mDocument.ModelSpace);
            var tCup = tCone.Cut2(tCutOut);

            tCone.Delete();
            tCutOut.Delete();

            tCup.Color = new EntityColor(System.Drawing.Color.Maroon);

            return tCup;
        }
    }
}
