﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using X3g.Plugin;

namespace SamplePlugin.CustomEntity
{
    class DrawGrid : ITool
    {
        public const string Key = "egr::SamplePlugin::DrawGrid";

        private IToolTemplate mTemplate;
        private bool mEnterFirstPoint;
        private GeVec3d mFirstPoint;
        private OverlayPolyline3d mRect;
        private DocumentManager mDocument;

        public DrawGrid(DocumentManager document)
        {
            mDocument = document;
        }

        public bool IsAvailable()
        {
            return (true);
        }

        public bool Start(String pArguments, IToolTemplate pTemplate)
        {
            mTemplate = pTemplate;
            mEnterFirstPoint = true;
            return (true);
        }

        public void Abort() { }
        public void Finish() { }

        public void StartPoint() { }

        public void FinishPoint()
        {
            if (mEnterFirstPoint)
            {
                mFirstPoint = mTemplate.GetPoint();
                mEnterFirstPoint = false;

                mRect = new OverlayPolyline3d(null);
                mRect.Color = System.Drawing.Color.Black;
                mRect.LineWidth = 2;
            }
            else
            {
                mRect.Dispose();
                mRect = null;

                CreateGrid(mFirstPoint, mTemplate.GetPoint());

                mTemplate.FinishTool();
            }
        }

        public void PointUpdated()
        {
            if (mRect != null)
            {
                GeVec3d start = mFirstPoint;
                GeVec3d end = mTemplate.GetPoint();

                double width = end.x - start.x;
                double height = end.y - start.y;

                GeVec3d[] vertices = {
                    start,
                    new GeVec3d(start.x + width, start.y, 0),
                    new GeVec3d(start.x + width, start.y + height,0),
                    new GeVec3d(start.x, start.y + height, 0),
                    start
                };

                mRect.SetVertices(vertices);
            }
        }

        private void CreateGrid(GeVec3d startPoint, GeVec3d endPoint)
        {
            GeBoundingBox3d bounds = new GeBoundingBox3d();
            bounds.Extend(startPoint);
            bounds.Extend(endPoint);

            Grid grid = new Grid(mDocument, bounds, 10, 10);
        }
    }
}
