﻿using SamplePlugin.CustomEntity;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using X3g.Plugin;

namespace SamplePlugin
{
    public class SamplePlugin : Plugin
    {
        const string SceneExplorerWindowName = "SceneExplorer";

        /// Matches <plugin hash=""> from plugin XML
        public override string X3gGetGlobalIdentifier() { return "11731D3B4E72A1B7BC51715212831A3E"; }
        
        /// Matches <plugin id=""> from plugin XML - module name used for plugin registration
        public override string X3gGetName() { return "SamplePlugin"; }
        
        /// Matches <plugin vendorName=""> from plugin XML - vendor name used for plugin registration
        public override string X3gGetVendor() { return "Sample Company"; }
        
        /// Obsolete: A more detailed an localized description can by given by <plugin\description lang="??"> in plugin XML
        public override string X3gGetDescription() { return "This is a sample plugin."; }
        
        /// Matches <plugin version=""> from plugin XML
        public override string X3gGetVersion() { return "1.1.0"; }
        
        public override bool X3gInitialize(string pPluginPath)
        {
            X3gActions.RegisterAction("egr::SamplePlugin::Load");
            X3gActions.RegisterAction("egr::SamplePlugin::Save");
            X3gActions.RegisterAction("egr::SamplePlugin::Export3ds");
            X3gActions.RegisterAction("egr::SamplePlugin::ExportDae");
            X3gActions.RegisterAction("egr::SamplePlugin::ExportFbx");
            X3gActions.RegisterAction("egr::SamplePlugin::ExportSkp");
            X3gActions.RegisterAction("egr::SamplePlugin::ExportDwg");
            X3gActions.RegisterAction("egr::SamplePlugin::ExportPec");
            X3gActions.RegisterAction("egr::SamplePlugin::ExportGltf");
            X3gActions.RegisterAction("egr::SamplePlugin::ExportUsdz");
            X3gActions.RegisterAction("egr::SamplePlugin::ExportCustom");
            X3gActions.RegisterAction("egr::SamplePlugin::CreateUserArticle");
            X3gActions.RegisterAction("egr::SamplePlugin::ListArticles");
            X3gActions.RegisterAction("egr::SamplePlugin::CustomDataAccess");
            X3gActions.RegisterAction("egr::SamplePlugin::ShowWindow");
            X3gActions.RegisterAction("egr::SamplePlugin::DiceTable");
            X3gActions.RegisterAction("egr::SamplePlugin::CreateRoom");
            X3gActions.RegisterAction("egr::SamplePlugin::AddCustomProperty");
            X3gActions.RegisterTool(DrawGrid.Key, new DrawGrid(Document));

            X3g.Plugin.MessageBox.CustomMessageBox = new CustomMessageBox();

            return (true);
        }

        public override void X3gFinalize()
        {
        }

        public override void X3gActivate()
        {
        }

        public override void X3gDeactivate()
        {
        }

        public override void X3gEvent(EventType pEventType, object pParameter)
        {
        }

        public override void X3gAction(string actionKey, string arguments)
        {
            if (actionKey == "egr::SamplePlugin::Load") Load();
            else if (actionKey == "egr::SamplePlugin::Save") Save();
            else if (actionKey == "egr::SamplePlugin::Export3ds") Export3ds();
            else if (actionKey == "egr::SamplePlugin::ExportDae") ExportDae();
            else if (actionKey == "egr::SamplePlugin::ExportFbx") ExportFbx();
            else if (actionKey == "egr::SamplePlugin::ExportSkp") ExportSkp();
            else if (actionKey == "egr::SamplePlugin::ExportDwg") ExportDwg();
            else if (actionKey == "egr::SamplePlugin::ExportPec") ExportPec();
            else if (actionKey == "egr::SamplePlugin::ExportGltf") ExportGltf();
            else if (actionKey == "egr::SamplePlugin::ExportUsdz") ExportUsdz();
            else if (actionKey == "egr::SamplePlugin::ExportCustom") ExportCustom();
            else if (actionKey == "egr::SamplePlugin::CreateUserArticle")
                CreateUserArticleInteractive(arguments);
            else if (actionKey == "egr::SamplePlugin::ListArticles") ListArticles();
            else if (actionKey == "egr::SamplePlugin::CustomDataAccess")
                InteractiveCustomData();
            else if (actionKey == "egr::SamplePlugin::ShowWindow") ShowWindow(arguments);
            else if (actionKey == "egr::SamplePlugin::DiceTable") CreateDiceTable();
            else if (actionKey == "egr::SamplePlugin::CreateRoom") CreateRoom();

            else if (actionKey == "egr::SamplePlugin::AddCustomProperty") AddCustomProperty();
        }

        private void AddCustomProperty()
        {
            if(this.Document.GetSelection().Count >= 1)
            {
                DbEntity selectedEntity = this.Document.GetSelection()[0];
                selectedEntity.GetPropertyProvider().Add(new CustomPropertyProvider());
                selectedEntity.GetPropertyProvider().Update(false);
            }
        }

        public override bool X3gSetWindowContent(Window pWindow)
        {
            if (pWindow.Name == SceneExplorerWindowName)
            {
                SceneExplorer.SceneExplorer explorer = 
                    new SceneExplorer.SceneExplorer(Document);
                SceneExplorer.SceneExplorerView view = 
                    new SceneExplorer.SceneExplorerView(explorer, Document);
                //view will be stretched to window client area
                //if width and height are NaN, otherwise width
                //and height would be fix
                view.Width = view.Height = double.NaN;
                pWindow.Content = view;
                pWindow.Title = I18N._("Scene Explorer");
                pWindow.SetInitialClientArea(300, 500);
                return (true);
            }
            else
            {
                return (false);
            }
        }

        /************** Load/Save Planning Functions ***************/

        private void Load()
        {
            string filename = @GetDesktopDir() + "/test.dwg";
            bool success = Document.LoadPlanning(filename);
            if (success)
            {
                X3g.Plugin.MessageBox.Show($"Successfully loaded {filename}");
            }
            else
            {
                X3g.Plugin.MessageBox.Show(
                    X3g.Plugin.MessageBox.Icon.Error,
                    $"Failed to load {filename}");
            }
        }

        private void Save()
        {
            string filename = @GetDesktopDir() + "/test.dwg";
            bool success = Document.SavePlanning(filename, DocumentFileFormat.Undefined, false);
            if (success)
            {
                X3g.Plugin.MessageBox.Show($"Successfully saved {filename}");
            }
            else
            {
                X3g.Plugin.MessageBox.Show(
                    X3g.Plugin.MessageBox.Icon.Error,
                    $"Failed to save {filename}");
            }
        }

        private string GetDesktopDir()
        {
            return Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
        }

        /************** Export Functions ***************/

        private void Export3ds()
        {
            var p = new Export3dsParams();
            p.AcisQuality = AcisQuality.Medium;
            p.CompatLevel = Export3DSCompatLevel.Meshes;
            p.DefaultEdgeSmoothness = 19.8;
            p.DefaultNormalSmoothness = 19.8;
            p.SingleSided = false;
            Geometry.Export3ds(GetDesktopDir() + "/Export3dsAll.3ds", p);
            p.Entities = GetEntityIds(Document.GetSelection());
            Geometry.Export3ds(GetDesktopDir() + "/Export3dsSel.3ds", p);
        }

        private void ExportDae()
        {
            var p = new ExportDaeParams();
            p.AcisQuality = AcisQuality.Medium;
            p.DefaultEdgeSmoothness = 19.8;
            p.DefaultNormalSmoothness = 19.8;
            p.Export2d = true;
            p.Export3d = true;
            p.Scale = 1.0;
            p.Unit = LengthUnit.Meter;
            p.DuplicateDoubleSided = true;
            p.MaxTextureSize = 4096;
            p.Edges = false;
            p.FlipMirroredFaces = false;
            Geometry.ExportDae(GetDesktopDir() + "/ExportDaeAll.dae", p);
            p.Entities = GetEntityIds(Document.GetSelection());
            Geometry.ExportDae(GetDesktopDir() + "/ExportDaeSel.dae", p);
        }

        private void ExportFbx()
        {
            var p = new ExportFbxParams();
            p.AcisQuality = AcisQuality.Medium;
            p.DefaultEdgeSmoothness = 19.8;
            p.DefaultNormalSmoothness = 19.8;
            p.SingleSided = false;
            p.Format = -1;
            p.EmbedMedia = true;
            p.WithLights = true;
            p.FlipMirroredFaces = false;
            Geometry.ExportFbx(GetDesktopDir() + "/ExportFbxAll.fbx", p);
            p.Entities = GetEntityIds(Document.GetSelection());
            Geometry.ExportFbx(GetDesktopDir() + "/ExportFbxSel.fbx", p);
        }

        private void ExportSkp()
        {
            var p = new ExportSkpParams();
            p.AcisQuality = AcisQuality.Medium;
            p.DefaultEdgeSmoothness = 19.8;
            p.DefaultNormalSmoothness = 19.8;
            p.ExportMaterials = true;
            p.FormatVersion = SkpFormatVersion.Default;
            Geometry.ExportSkp(GetDesktopDir() + "/ExportSkpAll.skp", p);
            p.Entities = GetEntityIds(Document.GetSelection());
            Geometry.ExportSkp(GetDesktopDir() + "/ExportSkpSel.skp", p);
        }

        private void ExportDwg()
        {
            Document.SavePlanning(@GetDesktopDir() + "/test.dwg", DocumentFileFormat.AutoCad2007, false);            
        }

        private void ExportPec()
        {
            using (FileStream pecFileStream = new FileStream(GetDesktopDir() + "/test.zip", FileMode.Create))
            {
                X3g.Plugin.IO.PecExport pecExport = new X3g.Plugin.IO.PecExport();
                pecExport.ExportDocument(pecFileStream);
                pecFileStream.Close();
            }
        }

        private void ExportGltf()
        {
            var fileDialog = new SaveFileDialog();
            fileDialog.Filter = "Binary GLTF |*.glb";
            fileDialog.DefaultExt = "glb";
            if (fileDialog.ShowDialog() == DialogResult.OK)
            {
                var gltfExport = new X3g.Plugin.IO.GltfExport();
                gltfExport.AcisQuality = AcisQuality.Medium;
                gltfExport.DuplicateDoubleSided = false;
                gltfExport.FormatVersion = X3g.Plugin.IO.GltfFormatVersion.Gltf2Binary;
                gltfExport.ExportDocument(fileDialog.FileName);
            }
        }

        private void ExportUsdz()
        {
            var fileDialog = new SaveFileDialog();
            fileDialog.Filter = "Usdz |*.usdz";
            fileDialog.DefaultExt = "usdz";
            if (fileDialog.ShowDialog() == DialogResult.OK)
            {
                var usdzExport = new X3g.Plugin.IO.UsdzExport();
                usdzExport.AcisQuality = AcisQuality.Medium;
                usdzExport.Scale = 1.0;
                usdzExport.ExportDocument(fileDialog.FileName);
            }
        }

        private void ExportCustom()
        {
            var fileDialog = new SaveFileDialog();
            fileDialog.Filter = "Custom XML Format |*.xml";
            fileDialog.DefaultExt = "xml";
            if (fileDialog.ShowDialog() == DialogResult.OK)
            {
                var entities = Document.ModelSpace.GetEntities();
                var components = Geometry.GetGeometrySeparated(
                    entities, GeometryCollectFlags.Collect2D3D);

                GeComponent root = new GeComponent();
                root.Instances = new List<GeComponentInstance>();
                for (int i = 0; i < components.Count; i++)
                {
                    var component = components[i];
                    component.Name = entities[i].Id;
                    var compInstance = new GeComponentInstance();
                    compInstance.Component = component;
                    compInstance.Layer = "";
                    compInstance.Transform = new GeMatrix();
                    root.Instances.Add(compInstance);
                }

                new CustomExport().Export(root, fileDialog.FileName);
            }
        }

        private void ShowWindow(string pWindowName)
        {
            if (string.IsNullOrEmpty(pWindowName))
                return;

            Window window = Windows.GetWindow(pWindowName);
            if (window == null)
                window = Windows.CreateDockableWindow(pWindowName);

            if (window != null)
                window.Visible = true;
        }

        private ReadOnlyCollection<string> GetEntityIds(IList<DbEntity> pEntities)
        {
            List<string> entityIds = new List<string>();
            if (pEntities != null)
            {
                foreach (var entity in pEntities)
                    entityIds.Add(entity.Id);
            }
            return entityIds.AsReadOnly();
        }

        /************** Create User Article Function **************/
        private void CreateUserArticleInteractive(string pArguments)
        {
            var tCreateUserArticle = new ArticleHelper.CreateUserArticle(Document, X3gArticles, X3gApp);

            if (pArguments == "selection")
                tCreateUserArticle.FromSelection();
            else
                tCreateUserArticle.Interactive();
        }

        /************** Fetch Articles and work with them *********/
        private void ListArticles()
        {
            var dialog = new ArticleHelper.FetchArticles(Document, X3gArticles);
            dialog.ShowDialog();
        }

        /************** Custom Data Access Functions **************/

        /// <summary>
        /// Will look for custom Entities and read the user defined CustomData Field
        /// </summary>
        private void InteractiveCustomData()
        {
            string tText;
            bool tDidSomething = false;
            foreach (DbEntity e in Document.GetSelection())
            {
                if (e.CustomData.Get("EG-GX-CustomData", out tText))
                {

                    X3g.Plugin.MessageBox.Show(
                        X3g.Plugin.MessageBox.Icon.Information,
                        tText);
                    tDidSomething = true;
                }
            }

            if (!tDidSomething)
            {
                X3g.Plugin.MessageBox.Show(
                    X3g.Plugin.MessageBox.Icon.Information,
                    I18N._("Please select an Article created by this Plugin"));
            }
        }

        private void CreateDiceTable()
        {
            var tHelpClass = new Create.Modeling(Document);
            tHelpClass.CreateDiceTable();
        }

        private void CreateRoom()
        {
            GeVec2d[] tCoordinates = {
                new GeVec2d(-5,-5),
                new GeVec2d(-5,5),
                new GeVec2d(5,5),
                new GeVec2d(5,-5),
            };

            // Create Floor and Ceiling
            {
                var tFloor = new X3g.Plugin.Room.Floor(tCoordinates);
                var tCeiling = new X3g.Plugin.Room.Ceiling(tCoordinates, 2.5);

                // Change properties of ceiling with propertyprovider
                DbEntity tEntity = DbEntity.FromId(tCeiling.Id);
                IPropertyProvider tPropertyProvider = tEntity.GetPropertyProvider();

                // Change ceiling type to grid
                IProperty tCeilingType = tPropertyProvider.GetProperty("x3g_room::CeilingGrid::ceilingtype");
                if (tCeilingType != null)
                    tCeilingType.SetValue(new ChoicePropertyValue("1"), true);
                else
                    Log.WriteToLog($"Missing Ceiling Property: \"x3g_room::CeilingGrid::ceilingtype\"");
            }

            // Create Walls
            X3g.Plugin.Room.Wall[] tWalls = {
                new X3g.Plugin.Room.Wall(tCoordinates[0], tCoordinates[1]),
                new X3g.Plugin.Room.Wall(tCoordinates[1], tCoordinates[2]),
                new X3g.Plugin.Room.Wall(tCoordinates[2], tCoordinates[3]),
                new X3g.Plugin.Room.Wall(tCoordinates[3], tCoordinates[0])
            };

            // Connect Walls
            tWalls[0].WallAtStart = tWalls[tWalls.Length - 1];
            foreach (int num in Enumerable.Range(1, tWalls.Length - 1))
                tWalls[num].WallAtStart = tWalls[num - 1];

            foreach (int num in Enumerable.Range(0, tWalls.Length))
                tWalls[num].WallAtEnd   = tWalls[(num + 1) % tWalls.Length];

            // Make Opening in Wall 0 of tWalls
            {
                var t = new X3g.Plugin.Room.Opening(tWalls[0], new GeVec3d(5, -3, 0.4));
                t.Flipped = true;
                var t2 = new X3g.Plugin.Room.Opening(tWalls[0], new GeVec3d(5, 3, 0.4));
                t2.Width = 3;
            }

            // Make Door with two Windows in Wall 2 of tWalls
            {
                var tDoor = new X3g.Plugin.Room.Door(tWalls[2], new GeVec3d(-5, 0, 0));
                tDoor.Flipped = true;

                DbEntity tEntity = DbEntity.FromId(tDoor.Id);
                IPropertyProvider tPropertyProvider = tEntity.GetPropertyProvider();
                if (tPropertyProvider == null)
                    return;

                IProperty tLeafType = tPropertyProvider.GetProperty("x3g_room::door::LeafType_0");
                if (tLeafType != null)
                    tLeafType.SetValue(new ChoicePropertyValue("2"), true);
                else
                    Log.WriteToLog($"Missing Door Property: \"x3g_room::door::LeafType_0\"");

                var tW = new X3g.Plugin.Room.Window(tWalls[2], new GeVec3d(-5, -2.5, 0.4));
                var tW2 = new X3g.Plugin.Room.Window(tWalls[2], new GeVec3d(-5, 1.5, 0.4));

                tW2.Position = new GeVec2d(-5, 2.5);
                tW2.Height = 1.2;
                tW2.Width = 2;
                tW2.Mirrored = true;
                tW2.Flipped = true;
                tW2.DistanceToFloor = 0.6;
            }

            // Make GlassPanel in Wall 3 of tWalls
            {
                var tGlassPanel = new X3g.Plugin.Room.GlassPanel(tWalls[3], new GeVec3d(0, 5, 0.4));
                tGlassPanel.Height = 2;
                tGlassPanel.Width = 8;

                DbEntity tEntity = DbEntity.FromId(tGlassPanel.Id);
                IPropertyProvider tPropertyProvider = tEntity.GetPropertyProvider();
                if (tPropertyProvider == null)
                    return;
                ReadOnlyCollection<IProperty> tPanelProperties = tPropertyProvider.GetProperties();
                foreach (IProperty prop in tPanelProperties)
                    Log.WriteToLog($"Panel Prop: {prop.Key}");

                IProperty tRows = tPropertyProvider.GetProperty("x3g_room::glasspanel::RowCount");
                if (tRows != null)
                    tRows.SetValue(new IntPropertyValue(2), true);
                else
                    Log.WriteToLog($"Missing glass panel Property: \"x3g_room::glasspanel::RowCount\"");

                IProperty tColumns = tPropertyProvider.GetProperty("x3g_room::glasspanel::ColumnCount");
                if (tColumns != null)
                    tColumns.SetValue(new IntPropertyValue(4), true);
                else
                    Log.WriteToLog($"Missing glass panel Property: \"x3g_room::glasspanel::ColumnCount\"");
            }

            // Create PolyWall with Floor, Door, Window, GlassPanel and Opening
            {
                GeVec2d[] tPolyPoints = { new GeVec2d(-4.9 , -5.1), new GeVec2d(-4.9, -14.9), new GeVec2d(4.9, -14.9), new GeVec2d(4.9 , -5.1),
                                      new GeVec2d(5.1, -5.1) , new GeVec2d(5.1, -15.1), new GeVec2d(-5.1, -15.1), new GeVec2d(-5.1, -5.1)};

                GeVec2d[] tPoints2 = { new GeVec2d(-5, -5), new GeVec2d(-5, -15), new GeVec2d(5, -15), new GeVec2d(5, -5) };
                var tFloor = new X3g.Plugin.Room.Floor(tPoints2);

                var tPolyWall = new X3g.Plugin.Room.PolyWall(tPolyPoints, 2.6);

                var tPolyDoor = new X3g.Plugin.Room.Door(tPolyWall, new GeVec3d(-5, -10, 0.4));
                var tPolyWindow = new X3g.Plugin.Room.Window(tPolyWall, new GeVec3d(5, -10, 0.4));
                var tPolyGlass = new X3g.Plugin.Room.GlassPanel(tPolyWall, new GeVec3d(0, -15, 0.4));
                tPolyGlass.Width = 4;
                tPolyGlass.Height = 2;
                var tOpening = new X3g.Plugin.Room.Opening(tPolyWall, new GeVec3d(-4, -15, 0.4));
            }
        }
    }
}
