﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using X3g.Plugin;

namespace SamplePlugin.SceneExplorer
{
    public class SceneExplorer
    {
        public SceneExplorer(DocumentManager pDocument)
        {
            mDocument = pDocument;
            mBoxes = new List<OverlayBoundingBox>();
        }

        public void BlockSelected(string pBlockName)
        {
            DisposeBoxes();

            try
            {
                Block block = mDocument.Blocks.GetBlock(pBlockName);
                if (block != null)
                {
                    //obtain block bounds by taking bounds from first reference
                    GeBoundingBox3d bounds = new GeBoundingBox3d();
                    var blockRefs = block.GetReferences(true);
                    foreach (var blockRef in blockRefs)
                    {
                        try
                        {
                            bounds = blockRef.GetBoundingBox(GeMatrix.Invert(blockRef.GetTransform()));
                        }
                        catch (System.Exception)
                        {
                            continue;
                        }
                        break;
                    }

                    var transforms = GetAllReferenceTransforms(block);
                    foreach (var transform in transforms)
                        mBoxes.Add(new OverlayBoundingBox(bounds, transform));
                }
            }
            catch (System.Exception e)
            {
                System.Console.WriteLine(e.ToString());
            }
        }

        public void EntitySelected(string pEntityId)
        {
            try
            {
                DbEntity entity = DbEntity.FromId(pEntityId);
                ShowEntityBounds(entity);
            }
            catch (System.Exception e)
            {
                System.Console.WriteLine(e.ToString());
            }
        }

        void ShowEntityBounds(DbEntity pEntity)
        {
            DisposeBoxes();

            GeMatrix blockTransform = GeMatrix.Identity;
            DbBlockReference blockRef = pEntity as DbBlockReference;
            if (blockRef != null)
            {
                blockTransform = blockRef.GetTransform();
            }

            GeBoundingBox3d bounds = pEntity.GetBoundingBox(GeMatrix.Invert(blockTransform));

            Block block = mDocument.Blocks.GetBlock(pEntity.OwningBlock);
            IList<GeMatrix> transforms = GetAllReferenceTransforms(block);
            foreach (var transform in transforms)
                mBoxes.Add(new OverlayBoundingBox(bounds, transform.PreMult(blockTransform)));
        }

        IList<GeMatrix> GetAllReferenceTransforms(Block pBlock)
        {
            List<GeMatrix> transforms = new List<GeMatrix>();

            if (pBlock.IsModelSpace)
            {
                transforms.Add(GeMatrix.Identity);
            } else
            {
                foreach (var blockRef in pBlock.GetReferences(true))
                {
                    IList<GeMatrix> refTransforms = GetAllReferenceTransforms(blockRef);
                    transforms.InsertRange(0, refTransforms);
                }
            }

            return (transforms);
        }

        IList<GeMatrix> GetAllReferenceTransforms(DbBlockReference pBlockRef)
        {
            List<GeMatrix> transforms = new List<GeMatrix>();

            Block block = mDocument.Blocks.GetBlock(pBlockRef.OwningBlock);
            IList<GeMatrix> parentTransforms = GetAllReferenceTransforms(block);
            foreach (var parentTransform in parentTransforms)
                transforms.Add(parentTransform.PreMult(pBlockRef.GetTransform()));

            return (transforms);
        }

        private void DisposeBoxes()
        {
            foreach(var box in mBoxes)
            {
                box.Dispose();
            }
            mBoxes.Clear();
        }

        DocumentManager mDocument;
        List<OverlayBoundingBox> mBoxes;
    }
}
